{-# LANGUAGE QuasiQuotes, OverloadedStrings, FlexibleContexts, UnicodeSyntax #-}
{-# OPTIONS_GHC -F -pgmF frquotes #-}
{-# OPTIONS -Wall -fno-warn-missing-signatures #-}

import Language.LaTeX
import qualified Language.LaTeX.Builder as B
import qualified Language.LaTeX.Builder.Internal as BI
import qualified Language.LaTeX.Builder.Beamer as BM
import qualified Language.LaTeX.Builder.Math as M
import qualified Language.LaTeX.Length as L
import Language.LaTeX.Builder.QQ
import Control.Monad.Writer
import Control.Applicative
import qualified Data.Generics.Uniplate.Data as U
import Data.Monoid.Unicode

import Data.List
import Data.Maybe

docName = "not-so-fresh-ml"

ø :: Monoid m => m
ø = (Data.Monoid.Unicode.∅)

-- this option should be used to print only the document body and not the
-- documentclass nor the preamble
withFmt = False -- True

main = quickView myViewOpts{basedir="out",pdflatex=pdflatex',showoutput=False,pdfviewer="echo"} docName doc
  where pdflatex' | withFmt   = "tex '&latex'"
                  | otherwise = pdflatex myViewOpts

doc = B.document dc preamb body where
  dc = BM.documentclass [BM.t,BM.red,BM.compress]

preamb =
     BI.usepackage ["utf8"] (BI.pkgName "inputenc")
  <> B.title «Not So Fresh ML»
  -- 'and' improve
  <> B.author «Nicolas Pouillard and François Pottier»
  <> B.institute (B.texttt "{Nicolas.Pouillard,Francois.Pottier}@inria.fr")
  <> BI.rawPreamble "\\date{CANS Seminar, 2009}"
  <> BM.beamertemplatenavigationsymbolsempty
  <> BM.useoutertheme [] "default"
  <> BM.useinnertheme [("shadow","true")] "rounded"
  <> BM.usecolortheme [] "orchid"
  <> BM.footline BM.defaultFootline{BM.showTotalFrames=False}
  <> margins

p = B.para

vcenter x = B.vfill <> x <> B.vfill

slide title = put . BM.slide title . vcenter . execWriter

put :: ParItem -> ParItemW
put = tell

alt x y = BM.only (BM.ovInts [1]) x <> BM.only (BM.ovInts [2]) y

margins = BM.setbeamersize (BM.TextMarginLeft (L.cm 0.3))
       <> BM.setbeamersize (BM.TextMarginRight (L.cm 0.3))

slideCs title subtitle =
    slide ø . put . B.center . (<>sub) . B.para . B.decls [B._Huge, B.bfseries] $ title
  where sub = mapNonEmpty ((B.vfill <>) . B.para) subtitle

slideC x = slideCs x ø

slideO title = put . BM.slideO title BM.ovIncr . vcenter . execWriter

visible = BM.visible . BM.ovInts . pure

box = BM.block

hs = M.protect -- M.verb
   . dropWhile (=='\n')

example :: LatexItem -> ParItemW
example = put . box ø . B.para

exampleHS :: String -> ParItemW
exampleHS = example . hs

exampleHS' :: String -> ParItemW
exampleHS' = put . B.para . hs

itemize = B.itemize
itemizeO = BM.itemize BM.ovIncr

item = tell . pure . B.item . B.para

notElm = M.parens . (M.notin <>)
{-
diff = M.parens . (M.neq <>)

symetricLink alpha atms beta =
  M.array [B.r, B.rtext ø, B.c, B.rtext ø, B.l]
          [B.cells [ alpha
                   , M.frac (M.decl M.scriptscriptstyle atms) (M.text (B.hspace (Em 3)))
                   , beta]
          ]
-}

link alpha atms beta =
  alpha <> M.stackrel (M.decl M.scriptscriptstyle atms) M.vartriangleright <> beta

subLink = M.sub (M.text $ M.verb "Link")
subAtom = M.sub (M.text $ M.verb "Atom")
idLink = B.math $ M.text (M.verb "id")<>subLink
compLink = B.math $ M.text (M.verb "(∘)") <> subLink
atomicLink = B.math $ M.text (M.verb "atomic") <> subLink
eqAtom = B.math $ M.text (M.verb "(≡)") <> subAtom
castAtom = B.math $ M.text (M.verb "cast") <> subAtom
castF = B.math $ M.text (M.verb "cast") <> M.sub (M.text $ M.verb "f")

-- style this
freshml = B.textsc «FreshML»
alphaCaml = «C» <> B.math M.alpha <> «ml»
purefreshml = «pure {freshml}»
-- ott = B.textsc «Ott»

alphaEquivalence = «{B.math M.alpha}-equivalence»

marknote = ParNote (TextNote "slicemark") ø
mark :: ParItemW
mark = put $ pure marknote

uncatPar (ParConcat pars) = pars
uncatPar par              = [par]

--slice = id
slice = fmap $ mconcat . comb [maySlice1, maySlice2] . uncatPar
  where notMark = (/= marknote)
        isMarked = any (== marknote) . U.universe
        maySlice1 = takeWhile notMark . drop 1 . dropWhile notMark
        maySlice2 = filter isMarked
        comb fs xs = fromMaybe xs . find (not . null) $ map ($xs) fs

body = slice $ execWriter $ do
  put B.maketitle

  slideC «Towards safer and more expressive languages for meta-programming»
  -- that's the direction we took with François for my Ph.D
  -- why two goals at once? because they are complementary...
  -- by meta-programming we means all programs that manipulate some kind of programs
  
  slideC «Program representation should stay well-typed and {BM.alert «well-scoped»}»
  -- the well-typed part guarantee that we can only build well-typed program
  --   representation, the benefit counter part is that you learn part of the
  --   type when pattern matching a program
  -- here we will omit the "well-typed" part since this talk is dedicated to
  --   bindings and scopes
  -- the well-scoped part guarantee that we don't generate programs with
  --   "unbound values", or "free variables"

  slide «Pursuing the work on {freshml}» $ do
    itemize !$? do
      item «Inspired from {freshml}»
      item «{purefreshml} for its safety»
      item «{alphaCaml} for its expressiveness»

  slideC «A taste of {freshml}/{alphaCaml}»

{-
  slide «Data type for explicitly typed lambda calculus» $ do

    exampleHS $
      [$istr|
      |data Term = Var Atom
      |          | App Term Term
      |          | Lam Ty   <Atom>Term
      |          | Let Term <Atom>Term
      |]

  slide «Computing the size of a term» $ do

    exampleHS $
      [$istr|
      |size :: Term → Int
      |size (Var _)      = 1
      |size (App t u)    = 1 + size t + size u
      |size (Lam _ <_>t) = 3 + size t
      |size (Let t <_>u) = 3 + size t + size u
      |]

  slide «» $ do

    exampleHS $
      [$istr|
        remove :: Atom → [Atom] → [Atom]
        remove _ [] = []
        remove a (b:bs)
          | a ≡ b     = remove a bs
          | otherwise = b : remove a bs

        fv :: Term → [Atom]
        fv (Var a)      = [a]
        fv (App t u)    = fv t ++ fv u
        fv (Lam _ <a>t) = remove a (fv t)
        fv (Let t <a>u) = fv t ++ remove a (fv u)
      |]
-}

  BM.frame (BM.ovInts [1]) ø [BM.label "dataTerm"]
           «Data type for explicitly typed lambda calculus» ø . vcenter !$? do

    -- * an atom is a name

    box ø . B.para . mconcat . zipWith (BM.only . BM.ovInts . pure) [1..] . map hs !$

      [ [$istr|
        |data Term
        |  = Var Atom
        |  | App Term Term
        |  | Lam < Atom × neutral Ty × inner Term >
        |  | Let < Atom × outer Term × inner Term >
        |]

      , [$istr|
        |data Term outer
        |  = Var (Atom outer)
        |  | App (Term outer) (Term outer)
        |  | ∃inner. Lam (Atom inner) Ty (Term inner)
        |  | ∃inner. Let (Atom inner) (Term outer) (Term inner)
        |]

      , [$istr|
        |data Term α
        |  = Var (Atom α)
        |  | App (Term α) (Term α)
        |  | ∃ β. Lam (Atom β) Ty (Term β)
        |  | ∃ β. Let (Atom β) (Term α) (Term β)
        |]

      , [$istr|
        |data Term α
        |  = Var (Atom α)
        |  | App (Term α) (Term α)
        |  | ∃ β. Lam (β ▹ α) (Atom β) Ty (Term β)
        |  | ∃ β. Let (β ▹ α) (Atom β) (Term α) (Term β)
        |]

      , [$istr|
        |data Term α
        |  = Var (Atom α)
        |  | App (Term α) (Term α)
        |  | Lam α<λβ→ (Ty, Term β)>
        |  | Let α<λβ→ (Term α, Term β)>
        |]
      ]

  slide «Capture avoiding substitution» $ do

    exampleHS $
      [$istr|
      |subst :: (Atom, Term) → Term → Term
      |subst (a, v) = go
      |  where
      |    go (Var b)       = if a ≡ b then v else Var b
      |    go (App t u)     = App (go t) (go u)
      |    go (Lam<b,ty,t>) = Lam<b, ty, go t>
      |    go (Let<b,t,u>)  = Let<b, go t, go u>
      |]

  slide «Computing the size of a term» $ do

    exampleHS $
      [$istr|
      |size :: Term → Int
      |size (Var _)      = 1
      |size (App t u)    = 1 + size t + size u
      |size (Lam<_,_,t>) = 3 + size t
      |size (Let<_,t,u>) = 3 + size t + size u
      |]

  slideC «{freshml} considered {visible 2 $ BM.alert «too fresh»}!»
    -- sometimes useless and thus inefficient

  slide «More efficient programs» $ do
    box «Freshening is useless while:» !$
      itemizeO $? do
        item «Computing the size of a term»
        item «Computing free variables»
        item «Typing some languages»
        item «Counting occurrences of some variable»
        item «Substituting closed terms for variables»
        item «Deciding {alphaEquivalence}»
        item «...»

  slide «To freshen or not to freshen?» $ do
    itemizeO !$? do
      item «While {freshml} implicitly freshen»
      item «This system allows both non-freshening and freshening openings»
      item «However we will only the non-freshening part»

  slideCs «World-index types for atoms» $
    visible 2 $ hs «let x = x in x»

  -- what is a world?

  slideO «Let's classify atoms by a world they live in» $ do
    box «The type of atoms is now indexed by a world» !$? do
      exampleHS' [$str|type Atom α|]
    box «Equality is homogeneous and prevents mixing worlds» !$
      p $ eqAtom <> hs [$str| :: ∀ α. Atom α → Atom α → Bool|]

  put $ BM.againframe (BM.ovInts [1,2,3]) ø [] "dataTerm"

  slideO «Worlds are closely related to each other» $ do
    {-
    box «Links between worlds» !$
      B.center $
      B.tabular (intersperse (B.rtext . B.hspace $ Em 4) [B.c,B.c,B.c]) [B.cells . map B.math $
        [ symetricLink M.alpha M.emptyset M.alpha
        , symetricLink M.alpha (diff M.a) M.beta
        , symetricLink M.alpha (notElm M._S) M.beta
        ]
      ]
    -}

    {-
    since (==) is polymorphic and homogeneous, it cannot return
    different results for different worlds.

    for any 'c', (|>c) is injective and so
      (x == y) == (x |> c == y |> c)
      if we can prove that x ∉ c and y ∉ c

    wrong x (<y>^c t) = x |> -c == y -- ?
                     == x |> -c |> c == y |> c -- mrong application of the thm
                     == x == y |> c -- which is wrong
    -}

    {-
    box «Links are oriented» !$
      B.center $
      B.tabular (intersperse (B.rtext . B.hspace $ Em 4) [B.c,B.c,B.c]) [B.cells . map B.math $
        [ link M.alpha M.emptyset M.alpha
        , link M.alpha (diff M.a) M.beta
        , link M.alpha (notElm M._S) M.beta
        ]
      ]
    -}
    {-
    One cannot use bidirectionally a link between two worlds.
    This because we must check before doing a cast that we are allowed to cast.
    To do this check we must compare atoms, since comparison is homogeneous
    then when doing the two checks one of them will be ill typed. It would
    make sense to have two version of the same atom that are known to be equal
    but with different worlds, because it would allow to tell if an atom is bound.
    -}
    box «The type of (oriented) links between worlds» !$? do
      exampleHS' [$str|type β ▹ α|]

    box «Links holds the set of atoms as a frontier» !$
      p . B.math $ link M.alpha (notElm M._S) M.beta

    box ø . B.center . p . BM.alert !$ «Links are supposed to be invisible/inferred!»

  put $ BM.againframe (BM.ovInts [3,4]) ø [] "dataTerm"

  slideO «Link operations» $ do

    box «Identity link» !$ p (idLink <> hs [$str| :: ∀ α. α ▹ α|])

    box «Link composition» !$
      p (compLink <> hs [$str| :: ∀ α β γ. (β ▹ γ) → (α ▹ β) → (α ▹ γ)|])

    box «Atomic link» !$
      p (atomicLink <> hs [$str| :: ∀ α. Atom α → (α ▹ α)|])

  slideO «Casts to walk through links» $ do
    box «Atomic casts» !$ p $ castAtom <> hs [$str| :: ∀ α β. (β ▹ α) → (Atom β → Atom α)|]

    box «Generalized casts» !$ p $ castF <> hs [$str| :: ∀ α β f. (β ▹ α) → (f β → f α)|]

    -- box «Swapped notation» !$ phs [$str|a ▹ lnk = cast lnk a|]

    -- casting an abstraction is just composing links
    -- castAbs lnk (Abs lnk' a t) = Abs (lnk . lnk') a t
    
    box ø . B.center . p . BM.alert !$ «Cast implies proof obligations or dynamic checks!»

  slideC «Atom abstraction as existential quantification»

  slide «Hiding the real world but keeping a link» $ do
    exampleHS [$str|data α<f> = ∃ β. Abs (β ▹ α) (Atom β) (f β)|]

  put $ BM.againframe (BM.ovInts [4,5]) ø [] "dataTerm"

  slideO «Making an abstraction» $ do
    exampleHS [$str|Abs :: ∀ α β f. (β ▹ α) → Atom β → f β → α<f>|]
    exampleHS [$str|Lam :: ∀ α β. (β ▹ α) → Atom β → Term β → Term α|]

    {-
    exampleHS [$istr|mkAbs :: ∀ α. Atom α → f α → α<f>
                    |mkAbs x t = Abs (atomic x) x t|]
    -}

    exampleHS [$istr|mkLam :: ∀ α. Atom α → Term α → Term α
                    |mkLam x t = Lam (atomic x) x t|]

    -- exampleHS [$str|nested x y = mkAbs x (mkAbs y (x, y))|]
    exampleHS [$str|mkConst x y = mkLam x (mkLam y (Var x))|]

  slideO «Opening an abstraction does not freshen it» $ do
    exampleHS [$str| let (Abs lnk x y) = t in u |]

    exampleHS [$str|Γ ⊢ t : α<f> where α ∈ Γ|]
    exampleHS [$str|Γ,β,lnk:β▹α,x:Atom β,y:f β ⊢ u : τ where β # τ|]

  -- slideC «Can we better understand data structures with binders?»
  -- tradionally FreshML freshen the binders when we open them
  -- by allowing to not freshen could help us to better understand
  -- what they mean "as is"

  slide «Safety Properties» $ do
    itemizeO !$? do
      item «Well-typed programs do not get stuck»
      item «{alphaEquivalence} is preserved by casts»
      item «Casts may dynamically fail or be proven successful»
      item «{alphaEquivalence} is defined structurally on types»

-- worlds as maps between actual atoms to fresh-denotations.

  slide «Example commuting abstraction with pairs» $ do
    exampleHS
      [$istr|
      |commute :: ∀ α. α<λβ→ (Term β, Term β)>
      |                  → (α<Term>, α<Term>)
      |commute t =
      |  let (Abs lnk x (y,z)) = t
      |  in (Abs lnk x y, Abs lnk x z)
      |]

  slide «Name capture does not type-check» $ do
    exampleHS
      [$istr|
      |wrong :: ∀ α. α<Term> → Term α → α<Term>
      |wrong t u =
      |  let (Abs lnk x y) = t
      |  in Abs lnk x u
      |]

  slideO «Computing the size of a term» $ do

    exampleHS $
      [$istr|
      |size :: ∀ α. Term α → Int
      |size (Var _)       = 1
      |size (App t u)     = 1 + size t + size u
      |size (Lam _ _ _ t) = 3 + size t
      |size (Let _ _ t u) = 3 + size t + size u
      |]

    box ø . B.center . p . BM.alert !$ «Polymorphic recursion!»

  slide «Computing free variables» $ do

    -- remove could also be a std fct

    example $ alt (hs
      [$istr|
      |remove :: Atom → [Atom] → [Atom]
      |remove _ [] = []
      |remove a (b:bs)
      |  | a ≡ b = remove a bs
      |  | otherwise = b : remove a bs
      |])

      (hs
      [$istr|remove ::
            |   ∀ β α. (β ▹ α) → Atom β → [Atom β] → [Atom α]
            |remove _   _ [] = []
            |remove lnk a (b:bs)
            |  | a ≡ b = remove lnk a bs
            |  | otherwise = cast lnk b : remove lnk a bs
            |])

    example $ alt (hs
      [$istr|
      |fv :: Term → [Atom]
      |fv (Var a)      = [a]
      |fv (App t u)    = fv t ++ fv u
      |fv (Lam<a,_,t>) = remove a (fv t)
      |fv (Let<a,t,u>) = fv t ++ remove a (fv u)
      |])

      (hs
      [$istr|
      |fv :: ∀ α. Term α →  [Atom α]
      |fv (Var a)         = [a]
      |fv (App t u)       = fv t ++ fv u
      |fv (Lam lnk a _ t) = remove lnk a (fv t)
      |fv (Let lnk a t u) = fv t ++ remove lnk a (fv u)
      |])

  slide «Looking up an environment» $ do

    exampleHS $
      [$istr|
      |data Env β = Empty
      |           | ∃ α. Snoc (β ▹ α) (Env α) (Atom β) Ty
      |]

    exampleHS
      [$istr|
      |lookupEnv :: ∀ α. Atom α → Env α → Ty
      |lookupEnv a (Snoc lnk env b ty)
      |  | a ≡ b    = ty
      |  | otherwise = lookupEnv (cast lnk a) env
      |lookupEnv _ Empty = error "unbound value"
      |]

  slide «Typing a term» $ do

    exampleHS $
      [$istr|
      |typing :: ∀ α. Env α → Term α → Ty
      |typing env (Var v)
      |  = lookupEnv v env
      |typing env (Lam lnk a ty t)
      |  = ty `TyArrow` typing (Snoc lnk env a ty) t
      |typing env (Let lnk a t u)
      |  = typing (Snoc lnk env a (typing env t)) u
      |typing env (App t u)
      |  = case typing env t of
      |      from `TyArrow` to | from ≡ typing env u → to
      |      _ → error "ill typed"
      |]

{-
  section «Status about the formalization»
    single abs, but arbitrary binding scope
    no ADTs sum/product
    some more typing rules
    reduction fails dynamically on casts, create, and fresh
    aeq
    thm
      item «{purefreshml} = this + single world»
-}

  slide «Challenges and future work» $ do
    itemizeO !$? do
      item «Deeper formalization and proofs»
      item «{alphaEquivalence} for inside-out abstractions»
      item «Better understanding of heterogeneous comparison»
      item «Integrating complex binding structures»
      item «Properties implied by world polymorphism»

  slide «Conclusion» $ do
    itemizeO !$? do
      item «Explicit scopes using world indices»
      item «Non-freshening opening»
      item «Atom abstraction as existential quantification»
      item «Expressivness close to a manual model with names»

  slideC «Questions?»

  put BM.appendix

  put $ BM.againframe ø ø [] "dataTerm"

  slideC «Polymorphic values represent closed terms»
  -- * forall alpha. Term alpha
  --   there cannot be any free atom in it
  -- * forall alpha. Term alpha -> Term alpha OR Atom alpha
  --   free atoms of the output are necessary a subset of input free atoms

  slideC «A more direct presentation of atom sorts»
  -- atom sorts are typically added afterward for pragmatic
  -- reasons, however they add some complexity to the presentation
  -- without adding any theoretical benefit.

  slideC «Generalizing {alphaCaml} data structures»
  -- * alphaCaml has two different kind of types,
  --   pattern types and expression types.
  -- * Pattern types describe the shape of binders.
  -- * Abstraction pack pattern types and expression
  --   types as a single abstract representation
  --   supporting alpha equivalence
  -- * These distinctions where guided by practical
  --   use cases rather than some uniform presentation.

  {-
  slideC «Reaching {ott} expressiveness and more»
  -- * Ott enables to define the set of bound atoms,
  --   and also where they are bound in a very general way.
  -- * This gives a more general presentation that patt/expr types
  -- * One would like to go further by allowing to dynamically
  --   select bound atoms.
  -}

  slideO «Picking fresh atoms» $ do
    example $ hs [$str|fresh x in t|]
           <> visible 3
                   (B.verb"           where x"<>"#"
                    <>B.math (M.parens(M.text(B.verb "t")<>M._Downarrow)))
    itemizeO !$? do
      item «The atom can be used in the world you like»
      item «Same proof obligation as in {purefreshml}»

  slideO «Picking fresh atoms (second version)» $ do
    exampleHS [$str|fresh x,lnkExp,lnkImp in t|]
    exampleHS [$istr|Γ,β,lnkExp:β▹α,lnkImp:α▹β,x:Atom β ⊢ t : τ
                    |   where α ∈ Γ, β # τ|]
    itemizeO !$? do
      item «The fresh atom is in an existential world»
      item «Links are provided to import and export things»
      item «Proof obligations relied to casts»

  slideO «Freshening is still available» $ do
    -- exampleHS [$str| let <fresh x>y = t in u |]
    exampleHS [$str| let (Abs _lnk (fresh x) y) = t in u |]
    box «Freshening allows to use the same world» !$? do
      exampleHS' [$str|Γ ⊢ t : α<f> where α ∈ Γ|]
      exampleHS' [$str|Γ,_lnk:α▹α,x:Atom α,y:f α ⊢ u : τ|]

  slide «Precise control over scopes» $ do
    box «Having explicit world subsume:» !$
      itemizeO $? do
        item «{alphaCaml} inner/outer/neutral annotations»
        item «{alphaCaml} pattern types/expression types distinction»
        item «{freshml}/{alphaCaml} atom sorts»

  slideC «Safe heterogeneous comparison!»

  slideO «Safe heterogeneous comparison!» $ do
    exampleHS [$str|atmEqH :: ∀ α β. Atom α → Atom β → (β ▹ α) → Bool|]
    example $ hs [$istr|atmEqH a b lnk | b ∉ lnk  = a ≡ cast lnk b
                       |               | otherwise = False|]

  slide «Substituting closed terms for variables» $ do
    exampleHS $
      [$istr|
      |substClosed ::
      |    ∀ α. Atm α → (∀ β. Term β) → Term α → Term α
      |substClosed a v = go id
      |  where
      |    go :: ∀ δ. (δ ▹ α) → Term δ → Term δ
      |    go lnk (Var b) | atmEqH a b lnk = v
      |                   | otherwise      = Var b
      |    go lnk (App t u)
      |      = App (go lnk t) (go lnk u)
      |    go lnk (Lam lnk' b ty t)
      |      = Lam lnk' b ty (go (lnk ∘ lnk') t)
      |    go lnk (Let lnk' b t u)
      |      = Let lnk' b (go lnk t) (go (lnk ∘ lnk') u)|]
      -- |    go _   (Cst cst)         = Cst cst

